import numpy as np
from faps.paternityArray import paternityArray
from faps.genotypeArray import genotypeArray
from faps.transition_probability import transition_probability
from faps.incompatibilities import incompatibilities
from warnings import warn

def paternity_array(offspring, mothers, males, mu=1e-12, missing_parents = 0, purge=None, selfing_rate=None, max_clashes=None, covariate=None):
    """
    Construct a paternityArray object for the offspring given known mothers
    and a set of candidate fathers using genotype data. Currently only SNP
    data is supported.

    Additional information about paternity from non-genetic sources can be
    provided through the argument `covariate`. For example, this might be a
    function of spatial distance between individuals, or social dominance.

    Parameters
    ---------
    offspring: genotypeArray, or dictionary of genotypeArrays
        Observed genotype data for one or more arrays of offspring in a
        single maternal family. To analyse multiple half-sib arrays in parallel,
        give a dictionary containing multiple named genotypeArray objects.
    mothers: genotypeArray, or dictionary of genotypeArrays
        Observed genotype data for the mother of the half-sib array. If multiple
        families of offspring are given, `mothers` should be a dictionary of
        genotypeArray objects with the same key names as for the offspring.
    males: genotypeArray
        Observed genotype data for the candidate males.
    mu: float between zero and one
        Point estimate of the genotyping error rate. Clustering is unstable if
        mu_input is set to exactly zero. Any zero values will therefore be set
        to a very small number close to zero (1e-12). Defaults to 1e-12
    missing_parents : float between zero and one
        Input value for the proportion of adults who are missing from the sample.
        This is used to weight the probabilties of paternity for each father
        relative to the probability that a father was not sampled. Defaults to
        zero, but you should definitely change this.
    purge: vector of str, or float between zero and one, optional
        Individuals who can be removed from the paternity array a priori. If
        a float is given, that proportion of individuals is removed from the
        array at random. Alternatively a string or vector of strings containing
        names of specific candidate fathers can be supplied.
    selfing_rate: float between zero and one, optional
        Input value for the prior probability of self-fertilisation.
    max_clashes: int, optional
        Maximum number of double-homozygous incompatibilities allowed between
        offspring and candidate fathers. Dyads with more incompatibilities than
        this will have their probability set to zero.
    covariate: 1-d array, or list of 1-d arrays, optional
        Vector of (log) probabilities of paternity based on non-genetic
        information, with one element for every candidate father. If this is a
        function of multiple sources they should be multiplied and included in
        this vector. If a dictionary for multiple half-sib arrays for have been
        supplied for offspring genotypes, `covariate` should likewise be a
        dictionary with and entry for each half-sib family and key names that
        match those given for the offspring.

    Returns
    -------
    A paternityArray, or a list of paternityArray objects.

    If covariate is not given this will be returned as a vector of zeros.

    Examples
    --------
    from faps import *
    import numpy as np

    # Generate a population of adults
    allele_freqs = np.random.uniform(0.3,0.5,50)
    adults = make_parents(20, allele_freqs)

    # Mate the first adult to the next three.
    mother = adults.subset(0)
    progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
    # Create paternityArray
    patlik = paternity_array(
        progeny,
        mother,
        adults,
        mu=0.0013,
        missing_parents=0.2
    )
    # View posterior probabilities of paternity (G matrix in the FAPS paper)
    patlik.prob_array()

    # Change some parameters that will affect the G matrix.
    patlik.missing_parents = 0.3
    patlik.selfing_rate = 0
    # Explicitly set some candidates to have probabilities of paternity to zero
    patlik.purge = 'base_0' # Another way to remove the mother
    patlik.purge = ['base_15', 'base_16', 'base_17'] # remove a list of specific candidates
    patlik.purge = 0.2 # Throw two candidates out at random

    # Another way to do the previous steps in a direct call to paternity_array.
    paternity_array(
        progeny,
        mother,
        adults,
        mu=0.0013,
        missing_parents=0.2,
        selfing_rate=0,
        purge = 0.2
    )
    """
    if mu == 0:
        mu = 1e-12
        warn('Setting error rate to exactly zero causes clustering to be unstable. mu set to 10e-12')

    # Check missing_parents
    if not isinstance(missing_parents, (int, float)):
        raise TypeError("missing_parents should be between 0 and 1.")
    if missing_parents < 0 or missing_parents >1:
        raise ValueError("missing_parents must be between 0 and 1!")
    if missing_parents ==1:
        warn("Missing_parents set to 100%. Are you sure this is what you mean?")
    # Check selfing_rate
    if selfing_rate is not None:
        if not isinstance(selfing_rate, (int, float)):
            raise TypeError("selfing_rate should be between 0 and 1.")
        if selfing_rate < 0 or selfing_rate >1:
            raise ValueError("selfing_rate must be between 0 and 1.")
        if selfing_rate == 1: warn("selfing_rate set to 100%. Are you sure that is what you meant?")
    # Check purge
    if purge is not None:
        if isinstance(purge, float):
            if purge <= 0 or purge >= 1:
                raise ValueError("purge must be between zero and one.")
        # If one or more integers is given, remove candidates at those indices
        elif isinstance(purge, (list, np.ndarray, int, str)):
            # If all entries are strings, find the names of the candidates.
            if all([isinstance(x, str) for x in purge]):
                if not all(np.atleast_1d(np.asarray(np.isin(purge, males.names)))):
                    raise ValueError("One or more names in paternityArray.purge are not found in paternityArray.candidates")
        else:
            raise TypeError("Error: purge should be a float or list of floats between zero and one.")

    #If a single halfsib family is given.
    if isinstance(offspring, genotypeArray) & isinstance(mothers, genotypeArray):
        # array of opposing homozygous genotypes.
        incomp = incompatibilities(males, offspring)
        # take the log of transition probabilities, and assign dropout_masks.
        prob_f, prob_a = transition_probability(offspring, mothers, males, mu)
        output = paternityArray(likelihood=prob_f,
                                lik_absent=prob_a,
                                offspring=offspring.names,
                                mothers=offspring.mothers,
                                fathers=offspring.fathers,
                                candidates=males.names,
                                mu=mu,
                                purge=purge,
                                missing_parents=missing_parents,
                                selfing_rate=selfing_rate,
                                clashes=incomp,
                                max_clashes=max_clashes)

        if covariate is not None:
            output.add_covariate(covariate)

        return output




    # If a list of genotype arrays for separate halfsib families are given:
    elif isinstance(offspring, list) & isinstance(mothers, list):
        if len(offspring) != len(mothers):
            raise ValueError('Lists of genotypeArrays are of different lengths.')

        # Set up input of covariates if necessary.
        if isinstance(covariate, list):
            if len(offspring) != len(covariate):
                raise ValueError("If a list of arrays of probabilities for covariates are supplied, this should have a row for every offspring genotypeArray.")
            cov = covariate
        elif covariate is None:
            cov = np.zeros(males.size * len(offspring)).reshape([len(offspring), males.size])
        else:
            raise TypeError("If covariates are supplied for multiple half-sib arrays, this should be a list of 1-d vectors, each with an element for each candidate.")


        output = [None] * len(offspring)
        for i in range(len(offspring)):
            # array of opposing homozygous genotypes.
            incomp = incompatibilities(males, offspring[i])
            # take the log of transition probabilities, and assign dropout_masks.
            prob_f, prob_a = transition_probability(offspring[i], mothers[i], males, mu)
            # create paternityArray and send to output
            patlik = paternityArray(likelihood=prob_f,
                                    lik_absent=prob_a,
                                    offspring=offspring[i].names,
                                    mothers=offspring[i].mothers,
                                    fathers=offspring[i].fathers,
                                    candidates=males.names,
                                    mu=mu,
                                    purge=purge,
                                    missing_parents=missing_parents,
                                    selfing_rate=selfing_rate,
                                    clashes=incomp,
                                    max_clashes=max_clashes)

            patlik.add_covariate(cov[i])
            output[i] = patlik


        return output

    # If a dictionary of genotype arrays for separate halfsib families are given:
    elif isinstance(offspring, dict) & isinstance(mothers, dict):
        if offspring.keys() != mothers.keys():
            raise ValueError('Dictionaries of genotypeArray objects for offspring and mothers do not match')

        # Set up input of covariates if necessary.
        if isinstance(covariate, dict):
            if offspring.keys() != covariate.keys():
                raise ValueError("Key names of genotypeArrays of offspring genotypes and covariates do not match.")
            cov = covariate
        elif covariate is None:
            # Create a dictionary of zero vectors.
            cov = {k: np.zeros(males.size) for k in offspring.keys()}
        else:
            raise TypeError("If offspring and maternal genotypes are supplied as dictionaries, covariates should either be a 1-d of log probabilities, or a dictionary with the same keys as those for genotypes.")

        output = {}
        for i in offspring.keys():
            # array of opposing homozygous genotypes.
            incomp = incompatibilities(males, offspring[i])
            # take the log of transition probabilities, and assign dropout_masks.
            prob_f, prob_a = transition_probability(offspring[i], mothers[i], males, mu)
            # create paternityArray and send to output
            patlik = paternityArray(likelihood=prob_f,
                                    lik_absent=prob_a,
                                    offspring=offspring[i].names,
                                    mothers=offspring[i].mothers,
                                    fathers=offspring[i].fathers,
                                    candidates=males.names,
                                    mu=mu,
                                    purge=purge,
                                    missing_parents=missing_parents,
                                    selfing_rate=selfing_rate,
                                    clashes=incomp,
                                    max_clashes=max_clashes)
            patlik.add_covariate(cov[i])
            output[i] = patlik
        return output

    else:
        raise TypeError("offspring and mothers should be genotypeArray objects for a single half-sib array and its mother, or else lists or dictionaries of half-sibships and mothers.")



import numpy as np

class matingEvents(object):
    """
    Information about a sample of possible of mating events drawn proportionally
    to their probability. These can then be used to infer mating patterns using
    and array of phenotypes for each male.

    Indices for possible fathers are drawn based on genetic data, other pertinent
    information (such as distance between parents), or a combination of these
    types. Mating events are drawn for 'units', which usually means partitions
    in a `sibshipCluster` object. These samples are then resampled in proportion
    to a weight on each unit. To assess uncertainty, the aggregate mating events
    can be subsampled.

    `matingEvent` objects are usually created from sibshipCluster objects using
    the fucntion `sibshipCluster.mating_event`. They could also be generated
    from a `paternityArray` if full sibship structure was not an issue, but this
    is not implented yet.

    Parameters
    ----------
    unit_names: array or list
        Names for each unit.
    candidates: array or list
        Names for each candidate.
    unit_weights: vector
        Relative contribution of each unit. If the matingEvent object is derived
        from a sibshipCluster object, this is the relative probability of each
        partition that had a non-zero probability.
    unit_events: array or list
        List of arrays, each of which documents mating events for a single unit.
    total_events: array
        Mating events for the whole sample. This is a resample of elements in
        each sublist of unit_events, weighted by the probability of each
        unit.
    subsamples: array, optional
        Array of subsamples of total_mating events.
    """

    def __init__(self, unit_names, candidates, unit_weights, unit_events, total_events, subsamples = None):
        self.unit_names   = unit_names
        self.candidates   = candidates
        self.unit_weights = unit_weights,
        self.unit_events  = unit_events
        self.total_events = total_events
        self.subsamples   = subsamples

    #if len(self.unit_weights) != len(self.unit_events):
    #    raise ValueError('Length of unit_weights ({}) does not match that of unit_events ({}).'.format(len(self.unit_weights), len(self.unit_events)))

    def draw_subsample(self, n_subsamples = 1000, subsample_size = None):
        """
        Draw many subsamples from `total_events` at random, with replacement.

        Parameters
        ----------
        n_subsamples: int
            Number of subsamples.
        subsample_size: int, optional
            Size fo each subsample. Defaults to 0.1 time the length of
            total_events.

        Returns
        -------
        An array of integers indexing sires.
        """
        if not isinstance(n_subsamples, int):
            raise ValueError('n_subsamples should be an integer.')
        if not isinstance(subsample_size, int) and subsample_size is not None:
            raise ValueError('subsample_size should be an integer.')
        if subsample_size is None:
            subsample_size = np.around(0.1*len(self.total_events)).astype('int')

        sub_events = np.random.choice(self.total_events, n_subsamples*subsample_size, replace=True)
        sub_events = sub_events.reshape([n_subsamples, subsample_size])

        return sub_events

    def as_trait(self, trait):
        """
        Recast a matingEvents object to return the phenotype values of the
        candidates drawn.

        Parameters
        ----------
        trait: 1-d array or dict
            Phenotypes are usually a 1-d vector of the phenotypes of each candidate
            father. Alternatively, if the phenotype depends on both the mother and the
            father, for example a Euclidean distance between the pair, then this can
            be supplied as a dictionary of 1-d vectors.

        Returns
        -------
        A matingEvents object with entries for unit_events, total_events and
        subsamples given as the phenotype of each candidate. Other attributes are
        unchanged.
        """
        if self.total_events.dtype != int:
            raise TypeError("Values in matingEvents should be integers indexing candidate fathers, but are type {}.".format(self.total_events.dtype))

        # For a vector of phenotypes with an entry for each mother.
        if isinstance(trait, np.ndarray):
            if trait.ndim != 1:
                raise ValueError("`trait` should be a 1-d vector, or else a dictionary, but has shape {}".format(trait.shape))
            output = matingEvents(
                unit_names   = self.unit_names,
                candidates   = self.candidates,
                unit_weights = self.unit_weights,
                unit_events  = {k: trait[v] for k,v in self.unit_events.items()},
                total_events = trait[self.total_events],
                subsamples   = trait[self.subsamples]
            )

        # For dictionaries of trait values with an key for each mother.
        elif isinstance(trait, dict):
            if not all([x in self.unit_events.keys() for x in trait.keys()]):
                raise ValueError("Not all keys for traits have a matching key in unit_events of the matingEvents object.")
            if not all([x in trait.keys() for x in self.unit_events.keys()]):
                raise ValueError("Not all keys for unit_events have a matching key in traits.")

            # Apply phenotypes to unit_events
            unit_events  = {k: trait[k][v] for k,v in self.unit_events.items()}
            # Resample total_events
            total_events = []
            for i in self.unit_events.keys():
                total_events = total_events + [np.random.choice(
                    a    = unit_events[i],
                    size = self.unit_weights[0][i], # For some reason, unit_weights is returned as a tuple. Index [0] allows for that.
                    replace=False
                )]
            # Flatten and coerce to array
            total_events = [x for y in total_events for x in y]
            total_events = np.array(total_events)

            # Recreate matingEvents object.
            output = matingEvents(
                unit_names   = self.unit_names,
                candidates   = self.candidates,
                unit_weights = self.unit_weights,
                unit_events  = unit_events,
                total_events = total_events,
                subsamples   = None
            )
            # Add subsamples
            output.subsamples = output.draw_subsample(
                n_subsamples=self.subsamples.shape[0],
                subsample_size=self.subsamples.shape[1]
            )
        else:
            raise TypeError("`trait` should be a 1-d vector, or else a dictionary.")

        return output

    # def reweight_units(self):
    #     """
    #     Draw a single subsample without replacement from each element in
    #     `unit_events` in proportion to its weight. This is equivalent to
    #     recreating `total_events`, but without concatenating the output.
    #
    #     Parameters
    #     ----------
    #     None.
    #
    #     Returns
    #     -------
    #     A dictionary of vectors indexing candidate fathers that were drawn as
    #     mates for each unit.
    #     """
    #     output = {}
    #     for i in self.unit_events.keys():
    #         output[i] = np.random.choice(
    #             a    = self.unit_events[i],
    #             size = self.unit_weights[0][i], # For some reason, unit_weights is returned as a tuple. Index [0] allows for that.
    #             replace=False
    #         )
    #     return(output)



import numpy as np
from faps.genotypeArray import genotypeArray
from faps.make_parents import make_parents
from faps.make_sibships import make_sibships
from faps.make_generation import make_generation
from time import time, localtime, asctime
from pandas import DataFrame as df
from sys import stdout
from tqdm import tqdm

def make_power(replicates, nloci, allele_freqs, candidates, sires, offspring, missing_loci, mu_real, mu_input=None, unsampled_real=None, unsampled_input=None, selfing_rate=0, cluster_draws=1000, exp_clusters=True, return_paternities=False, return_clusters=False, verbose=True, progress=True):
    """
    Perform replicate simulation of half sibling arrays, and return accuracies about
    sibship reconstruction and posterior probabilties of paternity.

    This is just calls `make_generation` many times for given values or sets of values
    of parameters. `make_power` checks the input and prints a summary back to the user.
    This can be turned off by setting `verbose` to False. By default the input and
    summary statistics are returned as a Pandas data frame. You can also return a
    list of the `sibshipCluster` objects generated by setting `return clusters` to True.

    Parameters
    ----------
    replicates: int
        Number of replicate simulations to run for each set of parameters.
    nloci: int or list
        Number of loci to simulate.
    allele_freqs: float, or list
        Population allele frequencies for generating reproductive adults.
        This can be a single float for uniform allele frequencies, a list of two
        floats indicating the upper and lower bounds to draw allele frequencies from
        a uniform distribution, or a list of frequencies for each of `nloci` loci.
        Values must be between 0 and 1.
    candidates: int or list
        Adult population size.
    sires: int or list
        Number of males to sire full sibships, or a list of males. Note that the mother
        is always the first candidate in the array, and is indexed as position zero.
    offspring: int or list
        Full sibship sizes. If a single integer has been passed to `sires`,
        this must be an integer indicating full sibship size, which is constant for
        all full sibships. If a list is passed to `sires`, this should be a list of
        the same length indicating the size of each individual full-sib family.
    missing_loci: float or list
        Real diploid allelic dropout rate. Float or list of floats. Values must be
        between 0 and 1.
    mu_real: float or list
        Real haploid genotyping error rate. Float or list of floats between 0 and 1.
        In the latter case, mu_input should be the same length as mu_real.
    mu_input: float or list
        The genotyping error rate used to construct the `paternityArray`. Can be
        a single float or a list of floats values between 0 and 1. In the latter case,
        mu_input should be the same length as mu_real.
    unsampled_real: float or list, optional
        Individuals to remove from the list of candidate males. If a float between
        zero and one is given, this indicates a proportion of candidates which will
        be removed at random, reflecting incomplete sampling. If a list of integers is
        given, this indicates specific indices of candidates to be removed, such as
        a known true sire, or the mother.
    unsampled_input: float or list, optional
        Value for the proportion of unsampled fathers to be used in sibship clustering.
        Can be a single float or a list of floats. Values must be between 0 and 1.
    selfing_rate: float or list, optional
        Value for the selfing rate to be used in sibship clustering. Can be a
        single float or a list of floats. Values must be between 0 and 1.
    cluster_draws: int or list, optional
        Number of Monte Carlo draws in sibship clustering.
    exp_clusters: logical, optional
        If `True` exponentiate pairwise sibship matrix to speed up sibship clustering
        with a potential cost to accuracy. See `sibship_clustering` for details.
    return_paternities: logical, optional
        If `True` the array of paternity probabilities object is returned for each
        simulated array.
    return_clusters: bool, optional
        If true, a `sibshipCluster` object is returned for each simulated array.
    verbose: bool, optional
        If true prints output on parameter input.
    progress: bool, optional
        If True, prints a progress bar to track simulations (only tested in Jupyter).
    Returns
    -------
    A data frame listing:
        0. simulation index
        1. number of loci
        2. allele frequencies
        3. number of candidates
        4. number of sires
        5. number of offspring
        6. genotype dropout rate
        7. real genotype error rate
        8. input genotype error rate
        9. proportion of missing candidates
        10. input value for the presumed number of missing fathers
        11. input selfing rate
        12. time in seconds to create paternityArray
        13. time in seconds to perform clustering
        14. binary indiciator for whether the true partition was included in the
            sample of partitions.
        15. difference in log likelihood for the maximum likelihood partition
            identified and the true partition. Positive values indicate that the
            ML partition had greater support.
        16. posterior probability of the true number of families.
        17. mean probabilities that a pair of full sibs are identified as full sibs.
        18. mean probabilities that a pair of half sibs are identified as half sibs.
        19. mean probabilities that a pair of half or full sibs are correctly
            assigned as such.
        20. mean probability of paternity of the true sires for those sires who
            had been sampled (who had non-zero probability in the paternityArray).
        21. mean probability that the sire had not been sampled for those
            individuals whose sire was truly absent (who had non-zero probability
            in the paternityArray).

    If `return_paternities` is True, the array of paternity probabilities object
    is returned for each simulated array. If `return_clusters` is true, a list is
    returned containing the data frame described nabove and a list of sibshipClusters.

    Examples
    --------
    import numpy as np

    make_power(allele_freqs, 100, [0,2,3], [5,4,3], mu_real=0.0015, mu_input = 0.0015, unsampled_input=0.05, selfing_rate = 0.3)
    """
    if isinstance(replicates, int) and replicates > 0:
        if verbose:
            print("{} of each parameter combination will be performed.".format(replicates))
    else:
        raise TypeError("R should be a positive integer giving the number of replicate simulations to be performed.")
        return None

    if isinstance(nloci, int):
        if verbose: print("Simulating {} diploid loci.".format(nloci))
        nloci = [nloci]
    elif isinstance(nloci, list) or isinstance(nloci, np.ndarray):
        if verbose:
            print("Simulating arrays with multiple number of loci: {}.".format(nloci))
    else:
        raise TypeError("nloci should be a positive integer, or a list of positive integer.")
    if any([nloci[i] < 0 or not isinstance(nloci[i], int) for i in range(len(nloci))]):
        raise TypeError("nloci should be a positive integer, or a list of positive integer.")
    stdout.flush()

    # set up allele frequencies
    if isinstance(allele_freqs, float):
        if allele_freqs <=0 or allele_freqs >=1:
            raise ValueError("Allele frequencies must be between 0 and 1.")
        elif verbose:
            print("Uniform minor allele frequency of {}.format(allele_freqs).")
    elif isinstance(allele_freqs, list) or isinstance(allele_freqs, np.ndarray):
        if any([allele_freqs[i] < 0 and allele_freqs[i] > 1 for i in range(len(allele_freqs))]):
            raise ValueError("Allele frequencies must be between 0 and 1.")
        if len(allele_freqs) == 2 and verbose:
                print("Drawing allele frequencies between {} and {}.".format(allele_freqs[0], allele_freqs[1]))
        elif len(allele_freqs) == nloci and nloci is not 2:
            allele_freqs = allele_freqs
            if verbose:
                print("Allele frequencies supplied by the user.")
    else:
        raise ValueError("Allele frequencies must be given as a single value between 0 and 1, a list of two elements, or else a vector of frequencies for each locus.")
    stdout.flush()

    # set up population parameters
    if isinstance(candidates, int):
        if verbose is True:
            print("Simulating adult populations of {} individuals.".format(candidates))
        candidates = [candidates]
    elif isinstance(candidates, list) or isinstance(candidates, np.ndarray):
        if verbose is True:
            print("Simulating adult populations of multiple sizes: {}.".format(candidates))
    else:
        raise TypeError("candidates should be an integer or list of integers.")
    stdout.flush()

    # Generating the progeny
    if isinstance(sires, int):
        if isinstance(offspring, int):
            if verbose is True:
                print("Simulating {} families of {} offspring.".format(sires, offspring))
        else:
            raise TypeError("If a single integer is given for sires, then offspring must be a single integer.")
    elif isinstance(sires, list) or isinstance(sires, np.ndarray):
        if isinstance(offspring, list) or isinstance(offspring, np.ndarray):
            if len(offspring) != len(sires):
                raise ValueError("offspring should be the same length as sires.")
            if verbose is True:
                print("Simulating {} full-sib families.".format(len(sires)))
        else:
            raise TypeError("sires and offspring should either both be an integer, or should be lists of identical length.")
    stdout.flush()

    # dropout rates
    if isinstance(missing_loci, list) or isinstance(missing_loci, np.ndarray):
        if verbose:
            print("Multiple real genotyping error rates: {}.".format(missing_loci))
    if isinstance(missing_loci, float) or missing_loci ==0:
        if verbose:
            print("{}% of per-locus genotypes will be removed at random.".format(100*missing_loci))
        missing_loci = [missing_loci]
    else:
        raise ValueError("missing_loci should be an float or list of floats between 0 and 1.")
    stdout.flush()

    # Real genotype error rates
    if isinstance(mu_real, list) or isinstance(mu_real, np.ndarray):
        if verbose:
            print("Multiple real genotyping error rates: {}.".format(mu_real))
    elif isinstance(mu_real, float) or mu_real == 0:
        if verbose:
            print("{}% of alleles will be mutated at random.".format(mu_real*100))
        mu_real = [mu_real]
    else:
        raise ValueError("mu_real should be an integer or list of integers.")
    if any([mu_real[i] < 0 or mu_real[i] > 1 for i in range(len(mu_real))]):
        raise ValueError("All values for mu_real should be between zero and one.")
    #if len(mu_real) > 1 and len(missing_parents) > 1 and len(mu_real) > 1 != len(missing_parents) > 1:
        #print("ERROR: If multiple values are given for mu_real and missing_parents supply the same number of arguments for each.
    stdout.flush()

    # input error rates
    if isinstance(mu_input, list) or isinstance(mu_input, np.ndarray):
        if verbose:
            print("Constructing paternity arrays using multiple input values for assumed genotype-error rate: {}.".format(mu_input))
    elif isinstance(mu_input, float) or mu_input==0:
        if verbose:
            print("Genotype error rate of {} will be used to construct paternity arrays.".format(mu_input))
        mu_input = [mu_input]
    elif mu_input is None:
        mu_input = mu_real
        if verbose is True:
            print("Input error rates taken as the real error rates.")
    else:
        print("mu_input should be a float or list of floats between zero and one, or else None.")
    if len(mu_real) > 1 and len(mu_input) > 1 and len(mu_real) != len(mu_input):
        raise TypeError("If multiple values are given for mu_real and mu_input supply the same number of arguments for each.")
    if any([mu_input[i] < 0 and mu_input[i] >= 1 for i in range(len(mu_input))]) and mu_input is not None:
        raise ValueError("All elements in mu_input must be greater or equal to zero and less than one.")

    stdout.flush()

    # remove superfluous individuals
    if unsampled_real == 0 or unsampled_real == 0.0 or unsampled_real is None:
        if verbose is True:
            print("No candidates to be removed.")
    elif isinstance(unsampled_real, float):
        if unsampled_real > 0 and unsampled_real <= 1:
            if verbose is True:
                print("Removing {}% of the candidates at random.".format(np.round(unsampled_real*100, 2)))
        else:
            raise ValueError("If unsampled_real is a float it should be greater or equal to zero and less than one.")
    elif isinstance(unsampled_real, list) or isinstance(unsampled_real, np.ndarray):
        if any([not isinstance(unsampled_real[i], int) or unsampled_real[i] < 0 for i in range(len(unsampled_real))]) :
            raise TypeError("If a list is given for unsampled_real, all values should be positive integers.")
        if verbose is True:
            print("Removing candidates in positions {}.".format(unsampled_real))
    else:
        raise TypeError("unsampled_real should either be a float between zero and one, or else a list indexing candidate individuals to unsampled_real.")
    stdout.flush()

    # input candidate sampling rate
    if unsampled_input is None:
        if unsampled_real == 0 or unsampled_real is [0.0]:
            if verbose is True:
                print("Constructing paternity arrays assuming complete sampling of candidates.")
            unsampled_input = [0.0]
        if isinstance(unsampled_real, float):
            unsampled_input = [unsampled_real]
            if verbose is True:
                print("Constructing paternity arrays assuming sampling proportion is known to be {}.".format(1-unsampled_input))
        if isinstance(unsampled_real, list) or isinstance(unsampled_real, np.ndarray):
            unsampled_input = [float(len(unsampled_real)) / candidates[i] for i in range(len(candidates))]
            if verbose is True:
                print("Constructing paternity arrays assuming {} candidates are missing.".format(len(unsampled_real)))
        else:
            unsampled_input = [unsampled_input]
            print("No prior parameter used for the proportion of missing candidates.")
    elif isinstance(unsampled_input, int) and unsampled_input == -9:
        unsampled_input = [None]
        if verbose:
            print("No prior parameter used for the proportion of missing candidates.")
    else:
        if isinstance(unsampled_input, list) or isinstance(unsampled_input, np.ndarray):
            if verbose is True:
                print("Constructing paternity arrays assuming multiple values of proportions of missing candidates: {}.".format(unsampled_input))
        elif isinstance(unsampled_input, float):
            if verbose is True:
                print("Proportion missing canidates set to {}.".format(unsampled_input))
            unsampled_input = [unsampled_input]
        else:
            raise TypeError("unsampled_input should be an float or list of floats between zero and one.")
        if any([unsampled_input[i] >1 or unsampled_input[i] < 0 for i in range(len(unsampled_input))]):
            raise TypeError("unsampled_input should be an float or list of floats between zero and one.")
    stdout.flush()

    # input selfing rate
    if isinstance(selfing_rate, float) or selfing_rate == 0:
        if verbose is True:
            print("Self-fertilisation rate of {}.".format(selfing_rate))
        selfing_rate = [selfing_rate]
    elif isinstance(selfing_rate, list) or isinstance(selfing_rate, np.ndarray):
        if verbose is True:
            print("Self-fertilisation rate of {}.".format(selfing_rate))
    else:
        raise TypeError("selfing_rate should be an float or list of floats between zero and one.")
        return None
    if any([selfing_rate[i] >1 or selfing_rate[i] < 0 for i in range(len(selfing_rate))]):
        raise ValueError("selfing_rate should be an float or list of floats between zero and one.")
        return None
    stdout.flush()

    # check sibship parameters
    if isinstance(cluster_draws, int):
        if verbose is True:
            print("Performing {} Monte Carlo draws for sibship inference.\n".format(cluster_draws))
    else:
        raise TypeError("Give an integer value for cluster_draws.")
    if not isinstance(exp_clusters, bool):
        raise TypeError("exp_clusters should be logical.")
        return None
    if not isinstance(verbose, bool):
        raise TypeError("exp_clusters should be logical.")
        return None
    stdout.flush()

    if verbose: print("Parameters set. Beginning simulations on {}.".format(asctime(localtime(time()) )))
    stdout.flush()
    t0 = time()
    nreps = len(nloci)*len(candidates)*len(mu_real)*len(missing_loci)*len(unsampled_input)*len(selfing_rate)*replicates
    results  = np.zeros([nreps, 22])
    patarrays= []
    clusters = []

    counter=0

    # # set up the progress bar, if required
    # if progress:
    #     from ipywidgets import FloatProgress
    #     from IPython.display import display
    #     fp = FloatProgress(min=0, max=nreps) # instantiate the bar
    #     display(fp) # display the bar

    for r in tqdm(range(replicates)):
        for l in nloci:
            for c in candidates:
                for mr in mu_real:
                    for ml in missing_loci:
                        for th in unsampled_input:
                            for sr in selfing_rate:
                                # set up allele frequencies
                                if isinstance(allele_freqs, float):
                                    af = np.repeat(allele_freqs, l)
                                elif len(allele_freqs) == 2:
                                    af = np.random.uniform(allele_freqs[0], allele_freqs[1], l)
                                elif isinstance(allele_freqs, list) or isinstance(allele_freqs, np.ndarray):
                                    af = allele_freqs
                                # run the simulation.
                                rx = make_generation(
                                    allele_freqs= af,
                                    candidates = c,
                                    sires = sires,
                                    offspring = offspring,
                                    missing_loci = ml,
                                    mu_real = mr,
                                    mu_input = mr,
                                    unsampled_real=unsampled_real,
                                    unsampled_input= th,
                                    selfing_rate=sr,
                                    cluster_draws = cluster_draws,
                                    exp_clusters = exp_clusters,
                                    return_paternities=True,
                                    return_clusters=True,
                                    counter=counter
                                )
                                # send output to tables
                                results[counter] = rx[0]
                                if return_paternities: patarrays = patarrays + [rx[1]]
                                if return_clusters:    clusters = clusters + [rx[2]]

                                counter += 1
                                # # update progress bar
                                # if progress:
                                #     fp.value += 1 # signal to increment the progress bar

    if verbose:
        print("Simulations completed after {} minutes.".format(round((time() - t0) / 60, 2)))

    cn =['rep','nloci','allele_freq','n_adults','n_sires','array_size',
         'missing_loci','mu_real','mu_input',
         'unsampled_real','unsampled_input','selfing_rate',
         'time_paternity','time_cluster',
         'partition_found','delta_loglik',
         'nfamilies',
         'acc_fs','acc_hs','acc_all',
         'prob_sires','prob_absent']
    if return_paternities or return_clusters:
        results = [df(results, columns=cn)]
        if return_paternities: results = results + [patarrays]
        if return_clusters:    results = results + [clusters]
        return results
    else: return df(results, columns=cn)



from distutils.core import setup

setup(name='faps',
      version='2.5.3',
      description=' Inference of paternity and sibling relationships accounting for uncertainty in genealogy',
      url='http://github.com/ellisztamas/faps',
      author='Tom Ellis',
      author_email='thomas.ellis@gmi.oeaw.ac.at',
      license='MIT',
      packages=['faps'],
      install_requires=['numpy','fastcluster','scipy', 'pandas'],
      include_package_data=True,
      zip_safe=False)



import numpy as np
import os

def export_to_colony(offspring, mothers, males, allele_freqs,
                     file_path, project_name, output_name, seed, dropout_rates, error_rates,
                     known_dads, known_mums, update_af = 0, two_sexes = 2, inbreeding = 0,
                     haplodiploid= 0, polygamy = [0,0], infer_clones = 0, size_scaling = 1,
                     sibship_prior   = [0,1.0,1.0], known_allele_fr = 1, nruns = 1,
                     run_length = 2, monitor_method  = 0, monitor_interval= 100000,
                     windows_gui = 0, analysis_method = 1, precision = 3, parents_present = [1,1],
                     n_incompatibilies = 3):
    """
    Export genotypeArray objects to run in Colony2.

    Variable descriptions are mostly copied from the manual for Colony2. See the Colony2
    manual for full descriptions of what they mean.

    The output is rather crude.

    Parameters
    ----------
    offspring, mothers, males: genotypeArray objects
        Genotype information on the the offspring, mothers, and
        candidate males.
    allele_freqs: array
        A vector of population allele frequencies.
    file_path: str
        Folder to store analyses
    project_name: str
        Dataset name.
    output_name: str
        Output file name
    seed: int
        Random seed for random number generator.
    dropout_rates: array-like
        List of dropout rates for each locus.
    error_rates: array-like
        List of error rates for each locus.
    known_dads, known_mums: lists
        Lists of names of the father and mother of each offspring individual.
        There should be a single entry for each offspring. If the parent
        is unknown, enter 'NA'.
    update_af: int
        0/1=Not updating/updating allele frequency
    two_sexes: int
        2/1=Dioecious/Monoecious species
    inbreeding: int
        0/1=No inbreeding/inbreeding
    haplodiploid: int
        0/1=Diploid species/HaploDiploid species
    polygamy: int
        0/1=Polygamy/Monogamy for males and females
    infer_clones: int
        0/1=Clone inference =No/Yes
    size_scaling: int
        0/1=Full sibship size scaling =No/Yes
    sibship_prior: int
        0,1,2,3=No,weak,medium,strong sibship size prior; mean paternal &
        maternal sibship size.
    known_allele_fr: int
        0/1=Unknown/Known population allele frequency
    nruns: int
        Number of runs
    run_length:
        1,2,3,4 =  short, medium, long, very long runs
    monitor_method: int
        0/1=Monitor method by Iterate#/Time in second
    monitor_interval: int
        Monitor interval in Iterate# / in seconds
    windows_gui: int
        0/1 running in Windows GUI version?
    analysis_method: int
        0/1/2 pairwise-likelihood score (PLS), full likelihood (FL), or
        the FL and PLS combined (FPLS) method.
    precision: int
        1/2/3=low/medium/high Precision for Fulllikelihood
    parents_present:int
        prob. of dad/mum included in the candidates

    Returns
    -------
    A .dat file suitable for using in Colony 2.
    """
    # Open a file
    fo = open(os.path.join(file_path, project_name) + '.dat', "wb")
    # write population parameters
    fo.write(project_name+' !Dataset name \n' +
             output_name +' !Output file name \n' +
             str(offspring.size)+' ! Number of offspring in the sample \n' +
             str(offspring.nloci)+' ! Number of loci \n'+
             str(seed)+' ! Seed for random number generator\n' +
             str(update_af)+' ! 0/1=Not updating/updating allele frequency \n' +
             str(two_sexes)+' ! 2/1=Dioecious/Monoecious species \n' +
             str(inbreeding)+' ! 0/1=No inbreeding/inbreeding \n'+
             str(haplodiploid)+' ! 0/1=Diploid species/HaploDiploid species \n' +
             str(polygamy[0])+'  '+str(polygamy[1]) + ' ! 0/1=Polygamy/Monogamy for males & females \n' +
             str(infer_clones)+' ! 0/1=Clone inference =No/Yes \n' +
             str(size_scaling)+' ! 0/1=Full sibship size scaling =No/Yes \n' +
             str(sibship_prior[0])+ ' ' + str(sibship_prior[1])+ ' ' + str(sibship_prior[2])+ ' ' + ' ! 0,1,2,3=No,weak,medium,strong sibship size prior; mean paternal & maternal sibship size \n' +
             str(known_allele_fr) + ' ! 0/1=Unknown/Known population allele frequency \n')

    # Number of alleles per locus
    a = ''
    for i in range(offspring.nloci):
        a += '2 '
    fo.write(a + ' !Number of alleles per locus \n')
    # Allele frequencies
    for l in range(offspring.nloci):
        fo.write(str(9) + '\t' + str(10) + '\n')
        fo.write(('%1.4f'%allele_freqs[l]) + '\t' + ('%1.4f'%(1-allele_freqs[l])) + '\n')
    fo.write('\n')

    # Analysis parameters
    fo.write(str(nruns)+' ! Number of runs \n' +
             str(run_length)+' ! Length of run \n' +
             str(monitor_method) + ' ! 0/1=Monitor method by Iterate#/Time in second \n' +
             str(monitor_interval) + ' ! Monitor interval in Iterate# / in seconds \n' +
             str(windows_gui) + ' ! non-Windows version \n' +
             str(analysis_method) + ' ! Give an indicated value of 0, 1 or 2 to tell Colony to use the pairwise- likelihood score (PLS), full likelihood (FL), or the FL and PLS combined (FPLS) method \n' +
             str(precision) + ' ! 1/2/3=low/medium/high Precision for Fulllikelihood \n' +
             '\n')

    # locus information and error rates
    t, c, d, m = '', '', '',''
    for l in range(offspring.nloci):
        t += 'mk'+str(l+1)+'\t'
        c += str(0) + '\t'
        d += ('%1.4f'%error_rates[l]) + '\t'
        m += ('%1.4f'%error_rates[l]) + '\t'
    fo.write(t + '\n' + # marker name
             c + '\n' + # dominant/codominant marker
             d + '\n' + # heterozygous allelic dropout rate
             m + '\n' + # other error rates
             '\n')

    # Offspring genotypes
    offspring.geno[np.isnan(offspring.geno)] = -9
    for o in range(offspring.size):
        thisrow = offspring.geno[o].flatten()+9
        t= offspring.names[o]
        for i in range(offspring.nloci*2):
            t = t + '\t' + str(thisrow[i])
        t = t + '\n'
        fo.write(t)

    # prob of including parents and numbers of mothers and fathers.
    if two_sexes is 1:
        fo.write('\n' +
             str(parents_present[0]) + '  ' + str(parents_present[1]) + ' ! prob. of dad/mum included in the candidates \n' +
             str(males.size)+ '  ' + str(males.size) + ' ! numbers of candidate males & females \n'+
             '\n')
    elif two_sexes is 2:
        fo.write('\n' +
             str(parents_present[0]) + '  ' + str(parents_present[1]) + ' ! prob. of dad/mum included in the candidates \n' +
             str(males.size)+ '  ' + str(mothers.size) + ' ! numbers of candidate males & females \n'+
             '\n')

    # Candidate male genotypes
    males.geno[np.isnan(males.geno)] = -9
    for m in range(males.size):
        thisrow = males.geno[m].flatten()+9
        t= males.names[m]
        for i in range(males.nloci*2):
            t = t + '\t' + str(thisrow[i])
        t = t + '\n'
        fo.write(t)
    fo.write('\n')

    # Maternal genotypes
    if two_sexes is 1: # Hermaphrodites
        for m in range(males.size):
            thisrow = males.geno[m].flatten()+9
            t= males.names[m]
            for i in range(males.nloci*2):
                t = t + '\t' + str(thisrow[i])
            t = t + '\n'
            fo.write(t)
    elif two_sexes is 2: # separate sexes
        mothers.geno[np.isnan(mothers.geno)] = -9
        for m in range(mothers.size):
            thisrow = mothers.geno[m].flatten()+9
            t= mothers.names[m]
            for i in range(mothers.nloci*2):
                t = t + '\t' + str(thisrow[i])
            t = t + '\n'
            fo.write(t)
    else:
        raise ValueError("two_sexes should be 1 for hermaphrodites or 2 for separate sexes.")


    fo.write('\n')

    # Number of offspring with a known father in the sample of fathers.
    if all(offspring.fathers == 'NA'):
        fo.write(str(0) + ' ' + str(n_incompatibilies) + ' ! Number of known father-offspring dyads \n\n')
    else:
        check_dads = offspring.fathers[offspring.fathers != 'NA']
        fo.write(str(sum([check_dads[i] in known_dads for i in range(offspring.size)])) + ' ' + str(n_incompatibilies) + ' ! Number of known father-offspring dyads \n')
        # Pairs of dyads of offspring and known fathers
        for o in range(offspring.size):
            if offspring.fathers[o] in known_dads:
                fo.write(offspring.names[o] + ' ' + known_dads[o] + '\n')
        fo.write('\n')

    # Number of offspring with a known mother in the sample of mothers.
    if all(offspring.mothers == 'NA'):
        fo.write(str(0) + ' ' + str(n_incompatibilies) + ' ! Number of known mother-offspring dyads \n\n')
    else:
        check_mums = offspring.mothers[offspring.mothers != 'NA']
        fo.write(str(sum([check_mums[i] in known_mums for i in range(offspring.size)])) + ' ' + str(n_incompatibilies) + ' ! Number of known mother-offspring dyads \n')
        # Pairs of dyads of offspring and known mothers
        for o in range(offspring.size):
            if offspring.mothers[o] in known_mums:
                fo.write(offspring.names[o] + ' ' + known_mums[o] + '\n')
        fo.write('\n')

    # Number of known paternal full sibships
    paternal_families = np.unique(known_dads)
    paternal_families = paternal_families[paternal_families != 'NA']
    fo.write(str(paternal_families.shape[0]) + ' ! Number of know  paternal full sibships \n')
    # individuals in each paternal full sibship
    for m in range(paternal_families.shape[0]):
        this_family =  offspring.names[known_dads == paternal_families[m]]
        f = ''
        for o in range(len(this_family)):
            f += this_family[o] + ' '
        fo.write(str(this_family.shape[0]) + ' ' + f + '\n')

    # number of known maternal full sibships
    fo.write('0' + ' ! Number of known maternal full sibships \n')

    # set number of excluded fathers and mothers to zero.
    fo.write('0' + ' ! Number of offspring with known excluded paternity \n')
    fo.write('0' + ' ! Number of offspring with known excluded maternity \n')

    # set number of excluded sibships to zero.
    fo.write('0' + ' ! Number of offspring with known excluded paternal sibships \n')
    fo.write('0' + ' ! Number of offspring with known excluded maternal sibships  \n')

    # Close opened file
    fo.close()



import numpy as np

def parent_index(offs_names, parent_names):
    """
    Finds the position of parental names in a vector of possible parental names.
    
    This is essentially a convenient wrapper for np.where().
    
    ARGUMENTS:
    offs_names: Either a string denoting the name of the parents to be located,
        or else a list of such strings.
        
    parent_names: 1-d array of parental names.
    
    RETURNS:
    A list of positions of the parental for each entry in offs_names.
    """
    if isinstance(offs_names, str):
        offs_names = [offs_names]
    return [np.where(parent_names == offs_names[x])[0][0] for x in range(len(offs_names))]


import numpy as np
from pandas import DataFrame
from faps.matingEvents import matingEvents

def mating_summary(mating_event1, mating_event2, phenotype, absent_label=np.nan, quantiles=[2.5, 9.75]):
    """
    Compare patterns of mating in two matingEvents objects, for example based on
    observed genetic data, and a null model. Returns the relative probability
    that any male of a given phenotype sired any offspring.Confidence intervals
    are derived from subsampling mating events in each object.

    At present only categorical phenotypes are supported. Because Numpy cannot
    handle vectors of strings that include possible NaN values well, categories
    should be represented by an integer labels.

    Parameters
    ----------
    mating_event1: matingEvents object
        Information on likley mating events under model 1, for example mating
        imferred from observed genetic data.
    mating_event2: matingEvents object
        Information on likley mating events under model 2, for example mating
        under a null model of equal siring probality, or mating in proportion
        to dispersal only.
    phenotype: array
        1-d numerical vector indicating phenotype category for each candidate.
    absent_label: float, int, nan
        Value to use for missing fathers. See documentation for matingEvents.
    percentiles = list
        List of two percentiles between 0 and 100 indicating upper and lower
        bounds for confidence intervals. Defaults to 95% intervals.

    Returns
    -------
    Dataframe listing:
        category: label for each category
        n: frequency of each category in the candidates
        model1: probability of mating with each category under mating model 1.
        model2: probability of mating with each category under mating model 1.
        ratio: ratio of model1 and model2
    Additional columns indicate confidence intervals around each estimate.
    """
    if mating_event1.subsamples.shape[0] != mating_event2.subsamples.shape[0]:
        raise ValueError('Arrays of subsamples are different sizes.')
    if not isintance(mating_event1, matingEvents):
        raise TypeError('mating_event1 and mating_events2 are not matingEvents objects')

    # phenotype vectors, not including NaN.
    dat1 = mating_event1.phenotype(phenotype, absent_label, remove_NaN=True)
    dat2 = mating_event2.phenotype(phenotype, absent_label, remove_NaN=True)
    # List of unique elements
    category = np.unique(np.append(dat1, dat2))
    # count each type
    count1 = [(dat1 == x).sum()      for x in category]
    count2 = [(dat2 == x).sum()      for x in category]
    base   = [(phenotype == x).sum() for x in category]
    # Take the average, included NaNs that were excluded
    count1 = np.array(count1).astype('float') / len(mating_event1.total_events)
    count2 = np.array(count2).astype('float') / len(mating_event2.total_events)
    # get relative probability
    ratio = count1 / count2

    vals = np.unique(phenotype)
    # phenotypes in each subsample
    subcounts1 = phenotype[mating_event1.subsamples.flatten()].values.reshape(mating_event1.subsamples.shape)
    subcounts1 = np.array([(subcounts1 == i).mean(1) for i in vals])

    subcounts2 = phenotype[mating_event2.subsamples.flatten()].values.reshape(mating_event2.subsamples.shape)
    subcounts2 = np.array([(subcounts2 == i).mean(1) for i in vals])

    # ratio among subsamples.
    sub_ratio = subcounts1 / subcounts2
    # confidence intervals.
    SE1 = np.percentile(subcounts1, q=quantiles, axis=1)
    SE2 = np.percentile(subcounts2, q=quantiles, axis=1)
    SEr = np.percentile(sub_ratio,  q=quantiles, axis=1)
    # output the summary statistics
    output = DataFrame({'category':category, 'n':base,
                  'model1':count1, 'lowerCI_1':SE1[0],  'upperCI_1':SE1[1],
                  'model2':count2, 'lowerCI_2':SE2[0],  'upperCI_2':SE2[1],
                  'ratio':ratio,   'lower_ratio':SEr[0],'upperCI_ratio':SEr[1]})
    # rearrange column order and return a dataframe.
    return output[['category','n',
                   'model1', 'lowerCI_1',   'upperCI_1',
                   'model2', 'lowerCI_2',   'upperCI_2',
                   'ratio',  'lower_ratio', 'upperCI_ratio']]



import numpy as np

def unique_rows(a):
    """
    Pull out the unique set of rows of an n*m matrix along axis n.
    
    Parameters
    ----------
    a: array
        An n*m matrix
    
    Returns
    -------
    A subset of one or more rows in a.
    """
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('', a.dtype)]*a.shape[1]))
    return unique_a.view(a.dtype).reshape((unique_a.shape[0], a.shape[1]))



import numpy as np
from faps.paternityArray import paternityArray
from faps.genotypeArray import genotypeArray

def transition_probability(offspring, mothers, males, mu, inverse=False):
    """
    Calculate per-locus transition probabilities given data on offspring, known
    mothers and candidate fathers. Also returns probabilities that paternal
    alleles are drawn from population allele frequencies.

    Parameters
    ----------
    offspring: genotypeArray, or list of genotypeArrays
        Observed genotype data for the offspring.
    mothers: genotypeArray, or list of genotypeArrays
        Observed genotype data for the offspring. Data on mothers need
        to be in the same order as those for the offspring.
    males: genotypeArray
        Observed genotype data for the candidate males.
    mu: float between zero and one
        Point estimate of the genotyping error rate. Clustering is unstable if
        mu_input is set to exactly zero.
    inverse: bool, optional
        If true, function return 1-transition probabilities, or the
        probability of *not* generating the offspring given maternal and
        candidate paternal genotypes

    Returns
    -------
    0. Array indexing offspring x candidates transition probabilities.
    1. Array indexing offspring only for transition probabilities
    from population allele frequencies.
    """
    if not isinstance(offspring, genotypeArray):
        raise TypeError('offspring is not a genotypeArray')
    if not isinstance(mothers, genotypeArray):
        raise TypeError('mothers is not a genotypeArray')
    if not isinstance(males, genotypeArray):
        raise TypeError('males is not a genotypeArray')

    # # If the genotypeArray does not include geno_probs, include this
    # if not hasattr(offspring, 'geno_probs'):
    #     offspring.geno_probs = offspring.calculate_geno_probs(mu = mu)
    # if not hasattr(mothers, 'geno_probs'):
    #     mothers.geno_probs   = mothers.calculate_geno_probs(mu = mu)
    # if not hasattr(males, 'geno_probs'):
    #     males.geno_probs     = males.calculate_geno_probs(mu = mu)

    # array of viable transition probabilities
    trans_prob_array = np.array([[[1,  0.5, 0  ],
                                  [0.5,0.25,0  ],
                                  [0,  0,   0  ]],
                                 [[0,  0.5, 1  ],
                                  [0.5,0.5, 0.5],
                                  [1,  0.5, 0  ]],
                                 [[0,  0,   0  ],
                                  [0,  0.25,0.5],
                                  [0,  0.5, 1  ]]])


    # Array of probabilities of drawing paternal allele if the dad is 0,1,2
    allele_freqs = males.allele_freqs()
    af = np.array([
        (1-allele_freqs)**2,
        allele_freqs * (1-allele_freqs),
        allele_freqs ** 2
    ])

    # Empty arrays to store probabilities of paternity, and of missing dads
    lik_array = np.zeros([offspring.size, males.size, offspring.nloci])
    lik_absent = np.zeros([offspring.size, offspring.nloci])

    geno = [0,1,2]
    for f in geno:
        for m in geno:
            for o in geno:
                # Multiply genotype probabilities by the transition probability
                lik_array  += \
                trans_prob_array[o,m,f] * \
                offspring.geno_probs[:,:,o][:, np.newaxis] * \
                mothers.  geno_probs[:,:,m][:, np.newaxis] * \
                males.    geno_probs[:,:,f][np.newaxis]
                # Multiply transition probs by prob of drawing paternal alleles at random
                lik_absent += \
                trans_prob_array[o,m,f] * \
                offspring.geno_probs[:,:,o] * \
                mothers.  geno_probs[:,:,m] * \
                af[f]

    # Sum log-likelihoods over (valid) loci and correct for dropouts..
    lik_array  = np.log(lik_array)
    lik_absent = np.log(lik_absent)

    lik_array = np.nanmean(lik_array, 2)  * offspring.nloci
    lik_absent = np.nanmean(lik_absent, 1) * offspring.nloci

    return [lik_array, lik_absent]
